#pragma once

#include <string>
#include <vector>

enum class TokenType {_while, comma, func, noteq, exit, int_lit, semi, open_paren, close_paren, open_curly, close_curly, ident, let,  _if, _else, eq, plus, plusplus, pluseq, minusminus, minuseq, star, sub, div, starstar, cmpr, gt, gteq, lt, lteq };

std::optional<int> bin_prec(TokenType type)
{
    switch (type) {
    case TokenType::sub:
        return 5;
    case TokenType::minusminus:
        return 1;
    case TokenType::minuseq:
        return 3;
    case TokenType::plus:
        return 4;
    case TokenType::plusplus:
        return 0;
    case TokenType::pluseq:
        return 2;
    case TokenType::div:
        return 6;
    case TokenType::star:
        return 7;
    case TokenType::starstar:
        return 8;
    case TokenType::cmpr:
        return 9;
    case TokenType::gt:
        return 10;
    case TokenType::gteq:
        return 11;
    case TokenType::lt:
        return 12;
    case TokenType::lteq:
        return 13;
    case TokenType::noteq:
        return 14;
    default:
        return {};
    }
}

struct Token {
    TokenType type;
    std::optional<std::string> value {};
};

class Tokenizer {
public:
    inline explicit Tokenizer(std::string src)
        : m_src(std::move(src))
    {
    }

    inline std::vector<Token> tokenize()
    {
        std::vector<Token> tokens;
        std::string buf;
        while (peek().has_value()) {
            if (std::isalpha(peek().value())) {
                buf.push_back(consume());
                while (peek().has_value() && std::isalnum(peek().value())) {
                    buf.push_back(consume());
                }
                if (buf == "exit") {
                    tokens.push_back({ .type = TokenType::exit });
                    buf.clear();
                }
                else if (buf == "let") {
                    tokens.push_back({ .type = TokenType::let });
                    buf.clear();
                }
                else if (buf == "if") {
                    tokens.push_back({ .type = TokenType::_if });
                    buf.clear();
                }
                else if (buf == "while") {
                    tokens.push_back({ .type = TokenType::_while });
                    buf.clear();
                }
                else if (buf == "else") {
                    tokens.push_back({ .type = TokenType::_else });
                    buf.clear();
                }
                else if (buf == "func") {
                    tokens.push_back({ .type = TokenType::func });
                    buf.clear();
                }
                else {
                    tokens.push_back({ .type = TokenType::ident, .value = buf });
                    buf.clear();
                }
            }
            else if (std::isdigit(peek().value())) {
                buf.push_back(consume());
                while (peek().has_value() && std::isdigit(peek().value())) {
                    buf.push_back(consume());
                }
                tokens.push_back({ .type = TokenType::int_lit, .value = buf });
                buf.clear();
            }
            else if (peek().value() == '(') {
                consume();
                tokens.push_back({ .type = TokenType::open_paren });
            }
            else if (peek().value() == ')') {
                consume();
                tokens.push_back({ .type = TokenType::close_paren });
            }
            else if (peek().value() == ',') {
                consume();
                tokens.push_back({ .type = TokenType::comma });
            }
            else if (peek().value() == '{') {
                consume();
                tokens.push_back({ .type = TokenType::open_curly });
            }
            else if (peek().value() == '}') {
                consume();
                tokens.push_back({ .type = TokenType::close_curly });
            }
            else if (peek().value() == ';') {
                consume();
                tokens.push_back({ .type = TokenType::semi });
            }
            else if (peek().value() == '!') {
                if (peek(1).value() == '=') {
                    consume();
                    consume();
                    tokens.push_back({ .type = TokenType::noteq });
                }
            }
            else if (peek().value() == '=') {
                if (peek(1).value() == '=') {
                    consume();
                    consume();
                    tokens.push_back({ .type = TokenType::cmpr });
                } else {
                    consume();
                    tokens.push_back({ .type = TokenType::eq });
                }
            }
            else if (peek().value() == '+') {
                if (peek(1).value() == '+') {
                    consume();
                    consume();
                    tokens.push_back({ .type = TokenType::plusplus });
                }
                else if (peek(1).value() == '=') {
                    consume();
                    consume();
                    tokens.push_back({ .type = TokenType::pluseq });
                } else {
                    consume();
                    tokens.push_back({ .type = TokenType::plus });
                }
            }
            else if (peek().value() == '*') {
                if (peek(1).value() == '*') {
                    consume();
                    consume();
                    tokens.push_back({ .type = TokenType::starstar });
                } else {
                    consume();
                    tokens.push_back({ .type = TokenType::star });
                }

            }
            else if (peek().value() == '-') {
                if (peek(1).value() == '-') {
                    consume();
                    consume();
                    tokens.push_back({ .type = TokenType::minusminus });
                }
                else if (peek(1).value() == '=') {
                    consume();
                    consume();
                    tokens.push_back({ .type = TokenType::minuseq });
                } else {
                    consume();
                    tokens.push_back({ .type = TokenType::sub });
                }
            }
            else if (peek().value() == '/') {
                consume();
                tokens.push_back({ .type = TokenType::div });
            }
            else if (peek().value() == '>') {

                if (peek(1).value() == '=') {
                    consume();
                    consume();
                    tokens.push_back({ .type = TokenType::gteq });
                } else {
                    consume();
                    tokens.push_back({ .type = TokenType::gt });
                }
            }
            else if (peek().value() == '<') {
                if (peek(1).value() == '=') {
                    consume();
                    consume();
                    tokens.push_back({ .type = TokenType::lteq });
                } else {
                    consume();
                    tokens.push_back({ .type = TokenType::lt });
                }
            }
            else if (std::isspace(peek().value())) {
                consume();
            }
            else {
                std::cerr << "You messed up!" << std::endl;
                exit(EXIT_FAILURE);
            }
        }
        m_index = 0;
        return tokens;
    }

private:
    [[nodiscard]] inline std::optional<char> peek(int offset = 0) const
    {
        if (m_index + offset >= m_src.length()) {
            return {};
        }
        else {
            return m_src.at(m_index + offset);
        }
    }

    inline char consume()
    {
        //std::cout << "Consumed: " << m_src.at(m_index) << std::endl;
        return m_src.at(m_index++);

    }

    const std::string m_src;
    size_t m_index = 0;
};